"""
Helper functions for the awdible package.
"""

import string


def asciize(txt: str) -> str:
    """Convert a string to ascii"""

    punct = list(string.punctuation)

    keep = "()[].-_"
    punct = [p for p in punct if p not in keep]

    for p in punct:
        txt = txt.replace(p, "")

    replace_dict = {
        "é": "e",
        "è": "e",
        "ê": "e",
        "à": "a",
        "â": "a",
        "î": "i",
        "ï": "i",
        "ô": "o",
        "ù": "u",
        "û": "u",
        "ç": "c",
        "œ": "oe",
        "æ": "ae",
        "'": "",
        "`": "",
    }

    for k, v in replace_dict.items():
        txt = txt.replace(k, v)

    for k, v in replace_dict.items():
        txt = txt.replace(k.upper(), v.upper())

    keep2 = "()[]-_"
    for k in keep2:
        txt = txt.replace(k, "_")

    txt = (
        txt.replace("__", "_")
        .replace("__", "_")
        .replace("__", "_")
        .replace("__", "_")
        .replace("__", "_")
    )

    txt = txt.strip()

    return txt



"""
Logger module to log the messages
"""

import logging
import os

# LOG_LEVEL = logging.INFO


def get_logger(name, log_folder="logs", log_level="WARNING"):
    """Get the logger."""

    # create formatter
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s [%(filename)s - %(module)s - %(lineno)s - %(funcName)s()]",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    # create logger
    logger = logging.getLogger(name)
    logger.setLevel(log_level)

    # create file handler
    fn = os.path.join(log_folder, "awdible.log")
    if not os.path.exists(log_folder):
        os.makedirs(log_folder)

    # add file handlers to the logger
    fileHandler = logging.FileHandler(fn)
    fileHandler.setFormatter(formatter)
    logger.addHandler(fileHandler)

    # add console handler to the logger
    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(formatter)
    logger.addHandler(consoleHandler)

    return logger


logger = get_logger(__name__)



"""
Console Application
"""

from cleo.application import Application

from awdible.console.gui import GuiCommand
from awdible.console.run import RunCommand

# from cleo.commands.command import Command
# from cleo.helpers import argument, option


# class GuiCommand(Command):
#     """GUI Command"""

#     name = "gui"
#     description = "Launch Streamlit local Web App someone"

#     def handle(self):

#         self.line("Launching Streamlit Web App...")


def main() -> int:
    application = Application()

    application.add(GuiCommand())
    application.add(RunCommand())

    exit_code: int = Application().run()
    return exit_code


if __name__ == "__main__":
    application = Application()

    application.add(GuiCommand())
    application.add(RunCommand())

    application.run()

    # main()



"""
Video module
"""

import os

from pytube import YouTube

from awdible.core.defaults import VIDEO_PREFIX
from awdible.logger import logger


class Video:
    """Get the video from Youtube, extract audio from the video and save it"""

    @classmethod
    def _get(
        cls,
        url: str,
        silent_mode=False,
    ) -> tuple:
        """Get the audio from the url with Youtube"""

        # MANAGE URL PROBLEM
        # TODO : better tot raise an error
        # This code should not be here
        if not url.startswith(VIDEO_PREFIX):
            url = VIDEO_PREFIX + url

        try:
            yt = YouTube(url)
            return yt

        except Exception as e:
            logger.critical(f"Error: {e}")
            logger.critical(f"url {url} not correct ")
            if not silent_mode:
                raise AttributeError(f"Video {url} not correct")

    @classmethod
    def _stream(
        cls,
        yt: YouTube,
        url: str,
        silent_mode=False,
    ) -> tuple:
        """Get the audio from the stream from the video"""

        try:
            media = yt.streams.filter(only_audio=True).first()
            title = yt.title.replace(" ", "_")
            return title, media

        except Exception as e:
            logger.critical(f"Error: {e}")
            logger.critical(f"Video {url} not found")
            if not silent_mode:
                raise FileNotFoundError(f"Video {url} not found")

    @classmethod
    def _save(
        cls,
        title: str,
        media: YouTube,
        dest: str,
        silent_mode=False,
    ) -> str:
        """Save the audio from the media"""

        # # this code sould not be here
        # if not os.path.exists(cls.DEST):
        #     os.makedirs(cls.DEST)

        #  download
        out = media.download(dest)

        # TODO : ascize the title
        # title = ascize(title)

        # addd the extension
        title += ".mp4"

        # rename the title
        fn = os.path.join(dest, title)
        os.rename(out, fn)

        return fn

    @classmethod
    def get_stream(cls, url, silent_mode=False):
        """Get and stream"""

        yt = cls._get(url, silent_mode=silent_mode)
        title, media = cls._stream(yt, url, silent_mode=silent_mode)

        return title, media

    @classmethod
    def get_stream_save(cls, url, dest, silent_mode=False):
        """Get and stream and save"""

        yt = cls._get(url, silent_mode=silent_mode)
        title, media = cls._stream(yt, url, silent_mode=silent_mode)
        fn = cls._save(title, media, dest, silent_mode=silent_mode)

        return fn



"""
Configurations for the app
"""

import os

import dotenv
from dotenv import dotenv_values

from .logger import logger


def get_config():
    """Get the config"""

    no_api_key = "No API Key sush as 'RAPID_API_KEY' found in the environment variables or in .env file"
    no_host_key = "No API Host as 'RAPID_API_HOST' found in the environment variables or in .env file"

    # Search for the API Key and Host in the environment variables
    env_key = os.getenv("RAPID_API_KEY", no_api_key)
    env_host = os.getenv("RAPID_API_HOST", no_host_key)

    if os.path.exists(".env/.env.prod"):
        values = dotenv_values(".env/.env.prod")
    elif os.path.exists(".env"):
        values = dotenv_values(".env")
    else:
        values = {}

    config = {
        "X-RapidAPI-Key": values.get("RAPID_API_KEY", env_key),
        "X-RapidAPI-Host": values.get("RAPID_API_HOST", env_host),
    }

    logger.critical(f"Config: {config.get('X-RapidAPI-Host', no_host_key)}")

    return config


config = get_config()



"""
Streamlit front-end for the Awdible app.
"""

import streamlit as st

from awdible.core.awdible import Awdible
from awdible.helpers import asciize

DEFAULT_VIDEO_URL = "https://www.youtube.com/watch?v=9diaThxYnKA"
DEFAULT_VIDEO_ID = "V62oKsHdsLU"
DEFAULT_VIDEO_QUERY = "jo l'rigolo"


def user_out(out, dest="./tmp/"):
    """User output"""

    out = out.replace(dest, "")
    out = asciize(out)
    out = out.removeprefix(".")

    return out


# header image
img = "./docs/assets/img/image.png"
st.image(img)  # caption="Thanks to DALL-E for the image ;) "


# title
st.title("Awdible - Just the best free version of audible")
st.write(
    """
Just paste the url of the video you want to download and click the button.
"""
)

# Text input
pos_url = DEFAULT_VIDEO_URL
url = st.text_input("Youtube url", pos_url)


# Button
if st.button("Submit"):
    # init the Awdible class
    awdible = Awdible(video=url, test_mode=True, streamlit=True)

    awdible.dest

    # do run
    outs = awdible.run()
    out = outs[0] if isinstance(outs, list) else outs

    # deliver the file
    with open(out, "rb") as f:
        st.download_button(
            "Download", f, file_name=user_out(out, awdible.dest)
        )  # mime="audio/mp4"



"""
GUI Command
"""

import os

import streamlit.web.cli as cli
from cleo.commands.command import Command
from cleo.helpers import argument, option

# from ..gui.front import front


class GuiCommand(Command):
    """GUI Command"""

    name = "gui"
    description = "Launch Streamlit local Web App someone"

    def handle(self):
        self.line("Launching Streamlit Web App...")

        # dirname = os.path.dirname(__file__)
        # filename = os.path.join(dirname, "awdible", "gui", "front.py")
        # args = []
        # cli._main_run(filename, args)

        os.system("streamlit run awdible/gui/front.py")



"""
Default values and constants.
"""

# Urls
VIDEO_PREFIX = "https://www.youtube.com/watch?v="
QUERY_PREFIX = "https://www.youtube.com/results?search_query="

# Videos
DEFAULT_VIDEO_URL = "https://www.youtube.com/watch?v=9diaThxYnKA"
DEFAULT_VIDEO_ID = "9diaThxYnKA"

# folders
DEFAULT_DEST = "./"
DEFAULT_TMP = "./tmp/"
DEFAULT_LOG = "./log/"
DEFAULT_FILE = None

# Audio output
DEFAULT_OUTPUT = "mp3"
OUTPUT_OPTIONS = ["mp3"]  # "wav", "ogg", "m4a", "flac", "opus"

# Crop
# TODO :Set as minutes ????
DEFAULT_CROP_LIMIT = 3600
CROP_LIMIT = [15 * 60, 3_600]

# TODO enable this as None and es
# TODO change as lang
DEFAULT_CONTEXT = "fr"
CONTEXT_OPTIONS = [
    "fr",
    "en",
    "es",
    None,
    "",
    "auto",
    "None",
    0,
    "0",
]


# Searc: perform an api search to find video ID
DEFAULT_SEARCH = False
DEFAULT_SEARCH_NUMBER = 1
MAX_SEARCH_NUMBER = 3

# DEFAULT_SEARCH_TOP_RESULTS = 1
# SEARCH_OPTIONS = [
#     True,
#     False,
#     "True",
#     "False",
#     "true",
#     "false",
#     0,
#     1,
#     "0",
#     "1",
# ]

# Force dowload even if ffmpeg is installed
DEFAULT_FORCE = False

# prefix
DEFAULT_PREFIX = False

# Asynchronous
DEFAULT_ASYNCHRONOUS = False

# Streamlit
DEFAULT_STREAMLIT = False
DEFAULT_PORT = 8501

# Test mode
DEFAULT_TEST_MODE = False

# API config
DEFAULT_CONFIG = {
    "X-RapidAPI-Key": "No API Key Found sush as 'X-RapidAPI-Key' found in the environment variables or in .env file",
    "X-RapidAPI-Host": "No API Host Found as 'X-RapidAPI-Host' found in the environment variables or in .env file",
}


DEFAULT_SLEEPER = 1.0
DEFAULT_SLEEPER_LONG = 3.0



"""
Awdible is a Python library for working as a Audible like.
"""

__VERSION__ = "0.2.5"
__DESCRIPTION__ = "Awdible - Just the best free version of audible"
__AUTHOR__ = "Alexandre Gazagnes"

# from .core import Awdible

# __all__ = ["Awdible"]



"""
Crop the audio
"""

from awdible.logger import logger


class Crop:
    """Crop the audio if needed"""

    pass

    # @classmethod
    # def cut(
    #     self,
    #     fn: str,
    #     start: str,
    #     end: str,
    #     n_item: int | None,
    #     total_items: int | None,
    # ) -> str:
    #     """Crop the audio"""

    #     # pattern
    #     replacer = (
    #         fn.replace(".mp3", "_cut.mp3")
    #         if not (n_item or total_items)
    #         else fn.replace(".mp3", f"_cut_{n_item}_{total_items}.mp3")
    #     )

    #     # update out
    #     out = fn.replace(".mp3", replacer)

    #     # split
    #     ffmpeg.input(fn, ss=start, to=end).output(out).run()

    #     return out

    # @classmethod
    # def split(self, fn: str, n_splits=4) -> list[str]:
    #     """Crop the audio and return the list of the fn"""

    #     # len video
    #     len_seconds = self.video_duration(fn)

    #     # compute chunks
    #     chunks = self._build_chunks(len_seconds, n_splits)
    #     n_chunks = len(chunks)

    #     # perfom the cuts
    #     fn_list = []
    #     for i, (start, end) in enumerate(chunks):
    #         fn_chunk = Crop.cut(fn, start, end, i, n_chunks)
    #         fn_list.append(fn_chunk)

    #     return fn_list

    # @classmethod
    # def _build_chunks(self, len_seconds: int, n_splits: int) -> list[tuple]:
    #     """Build the chunks"""

    #     chunks = []

    #     # compute chunks
    #     for i in range(n_splits):
    #         start = i * len_seconds // n_splits
    #         end = (i + 1) * len_seconds // n_splits
    #         chunks.append((start, end))

    #     # -1 sec for each start except the first
    #     for i, (start, end) in enumerate(chunks):
    #         if not i:
    #             continue
    #         start -= 1
    #         chunks[i] = (start, end)

    #     # +1 sec for the last end
    #     chunks[-1] = (chunks[-1][0], chunks[-1][1] + 2)

    #     return chunks

    # @classmethod
    # def video_duration(self, fn):
    #     """Compute the video len"""

    #     len_seconds = ffmpeg.probe(fn)["format"]["duration"]
    #     len_seconds = int(float(len_seconds)) + 1

    #     return len_seconds



"""
Run command
"""

import os

from cleo.commands.command import Command
from cleo.helpers import argument, option

from awdible.core.awdible import Awdible
from awdible.logger import logger


class RunCommand(Command):
    """Run Awdible session"""

    name = "run"

    description = "Run Awdible session"

    arguments = [
        argument(
            "video",
            description="The video url to download from youtube. \nIf you want to input a search query, use the --search option",
            optional=True,
            default=Awdible.DEFAULT_VIDEO_URL,
        )
    ]
    options = [
        option(
            "dest",
            "d",
            description="The destination directory of the file",
            flag=False,
            default=Awdible.DEFAULT_DEST,
        ),
        option(
            "file",
            "f",
            description="The input file With list of urls/videos",
            flag=False,
        ),
        option(
            "output",
            "o",
            description="The output file name",
            flag=False,
            default=Awdible.DEFAULT_OUTPUT,
        ),
        option(
            "crop_limit",
            "l",
            description="The limit of duration of the video to download. If the video is longer than the limit, it will be cropped into smaller parts.",
            flag=False,
            default=Awdible.DEFAULT_CROP_LIMIT,
        ),
        option(
            "context",
            "c",
            description="The context of the search query. Can be 'fr' or 'en'",
            flag=False,
            default=Awdible.DEFAULT_CONTEXT,
        ),
        option(
            "search",
            "s",
            description="If set, the video argument will be treated as a search query and the first result will be downloaded.",
            flag=True,
        ),
        option(
            "prefix",
            "p",
            description=f"If set, can ommit the url prefix '{Awdible.VIDEO_PREFIX}' when passing the video url.",
            flag=True,
        ),
        option(
            "asynchronous",
            "a",
            description="If set, the download will be asynchronous.",
            # This means that the download will be done in the background and the command will return immediately. The download progress can be checked using the `awdible progress` command.
            flag=True,
        ),
    ]

    def handle(self):
        """handle the command"""

        # arguments
        video = self.argument("video")

        # no flags options
        dest = self.option("dest")
        file = self.option("file")
        output = self.option("output")
        crop_limit = self.option("crop_limit")
        context = self.option("context")

        # flags options
        search = self.option("search")
        asynchronous = self.option("asynchronous")
        prefix = self.option("prefix")

        # useless logging
        logger.debug(f"video: {video}")

        logger.debug(f"dest: {dest}")
        logger.debug(f"file: {file}")
        logger.debug(f"output: {output}")
        logger.debug(f"crop_limit: {crop_limit}")
        logger.debug(f"context: {context}")

        logger.debug(f"search: {search}")
        logger.debug(f"prefix: {prefix}")
        logger.debug(f"asynchronous: {asynchronous}")

        # check context
        if context not in ["fr", "en"]:
            raise Exception("Invalid context. Should be 'fr' or 'en'")

        # check if the dest is a valid directory
        if dest:
            if not os.path.isdir(dest):
                raise Exception("Invalid directory path")

        # check if the file is a valid file
        if file:
            if not os.path.isfile(file):
                raise Exception("Invalid file path")

        # check if the output is a valid file format
        if output not in ["mp3", "mp4"]:
            raise Exception("Invalid output file format : Only supports mp3 and mp4")

        # check if the crop limit is a valid number
        try:
            crop_limit = int(crop_limit)
        except ValueError:
            raise Exception("Invalid crop limit. Should be a number")
        if crop_limit < 0 or crop_limit > 3600 * 10:
            raise Exception(
                "Invalid crop limit. Should be between 0 and 36 000 seconds (10 hours)"
            )

        # check incompatible options video args and file otions are not passed together
        # unless default video is passed
        if (video and file) and (video != Awdible.DEFAULT_VIDEO_URL):
            raise AttributeError(
                "You can only pass either a video url or a file with list of urls"
            )

        # if file: make a liste of videos
        if file:
            with open(file, "r") as f:
                video_list = f.readlines()

        else:
            video_list = video.split(",")

        # strip items in the list
        video_list = [v.strip() for v in video_list]
        video_list = [v for v in video_list if v]
        video_list = [v for v in video_list if not v.startswith("#")]

        # # if prefix is set, add the prefix to the video list
        # if prefix:
        #     video_list = [Awdible.VIDEO_PREFIX + v for v in video_list]

        # # check video list
        # for v in video_list:
        #     if not v.startswith(Awdible.VIDEO_PREFIX):
        #         raise AttributeError(
        #             f"Invalid youtube video url. Should be in the format '{Awdible.VIDEO_PREFIX}', recieved : '{v}'\n Consier using the --search option to search for the video url or add the --prefix option to add the prefix to the video url."
        #         )

        self.line("Eh! I'm running the command")
        Awdible(
            video=video_list,
            dest=dest,
            file=file,
            output=output,
            context=context,
            prefix=prefix,
            search=search,
            asynchronous=asynchronous,
            crop_limit=crop_limit,
            streamlit=False,
        ).run()



"""
Module for validators and decorators 
"""

import os

# import logging
from abc import ABC, abstractmethod

from awdible.logger import logger


class Validator(ABC):
    """Abstract class for validators"""

    def __set_name__(self, owner, name):
        self.private_name = "_" + name

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        self.validate(value)
        setattr(obj, self.private_name, value)

    @abstractmethod
    def validate(self, value):
        pass



"""
Search module
"""

import requests

from awdible.core.defaults import QUERY_PREFIX, VIDEO_PREFIX
from awdible.logger import logger


class Search:
    """Search for the video with a specific query"""

    QUERY_PREFIX = QUERY_PREFIX
    VIDEO_PREFIX = VIDEO_PREFIX

    @classmethod
    def _find(
        cls,
        keywords: str,
        config: dict,
        context: str = None,
        lang: str = "NotImplemented",
        silent_mode: bool = False,
    ) -> list[str]:
        """Get the list of videos for a specific query"""

        # params
        params = {"q": keywords}
        # NOTE : is it a good thing to udpate keywords ???
        logger.info(f"searching for  params {params}")

        # context as lang for now, netx will be different
        context = context.lower().strip()[:2]
        if "fr" in context:
            params["hl"], params["gl"] = "fr", "FR"
        elif "en" in context:
            params["hl"], params["gl"] = "en", "US"
        elif "us" in context:
            params["hl"], params["gl"] = "en", "US"
        elif "uk" in context:
            params["hl"], params["gl"] = "en", "GB"
        elif "es" in context:
            params["hl"], params["gl"] = "es", "ES"
        else:
            pass

        # log params
        logger.info(f"searching for  params {params}")

        if not config:
            config = {}
        # manage keys if needed
        if config.get("X-RapidAPI-Key", None) is None:
            if not silent_mode:
                raise ValueError(
                    f"No API Key found in the environment variables : {config.get('X-RapidAPI-Key', None)}"
                )

        if config.get("X-RapidAPI-Host", None) is None:
            if not silent_mode:
                raise ValueError(
                    f"No API Host found in the environment variables : {config.get('X-RapidAPI-Host', None)}"
                )

        if config.get("X-RapidAPI-Host", None) != "youtube-data8.p.rapidapi.com":
            if not silent_mode:
                raise ValueError(
                    f"No API Host found in the environment variables : {config.get('X-RapidAPI-Host', None)}"
                )

        url = "https://youtube-data8.p.rapidapi.com/search/"
        response = requests.get(url, headers=config, params=params)

        # raise for status
        # TODO : manage side effect
        if not silent_mode:
            response.raise_for_status()  # raise an error if the request fails

        json = response.json()

        return json

    @classmethod
    def _parse(
        cls,
        json: str,
        n_results=5,
        silent_mode=False,
    ) -> list[str]:
        """Parse the json response to get the video id and title"""

        contents = json["contents"]

        contents = contents[:n_results]

        # if contents are not found
        if not len(contents):
            logger.error(f"No video found in parsing the json : contents = {contents}")
            logger.error(json)
            if not silent_mode:
                raise ValueError(
                    f"No video found in parsing the json : contents = {contents}"
                )

        logger.info(contents)

        # extract video
        try:
            videos = [x.get("video", None) for x in contents]
        except AttributeError as e:
            logger.error(f"Error in parsing the json with get('video') : {e}")
            # logger.error(f"contents is : {contents}")
            if not silent_mode:
                raise e

        # extract video id and title
        try:
            li = [(x["title"], x["videoId"]) for x in videos]
            logger.info(li)

        except Exception as e:
            logger.error(f"Error in parsing the json with 'title' and 'ID') : {e}")
            # logger.error(f"contents is : {contents}")
            if not silent_mode:
                raise e

        # filter id
        # get video id
        videos = [x["videoId"] for x in videos]

        # make final url
        VIDEO_PREFIX = (
            cls.VIDEO_PREFIX
            if cls.VIDEO_PREFIX.endswith("/")
            else cls.VIDEO_PREFIX + "/"
        )
        videos = [VIDEO_PREFIX + x for x in videos]

        return videos

    @classmethod
    def find_parse(
        cls,
        keywords: str,
        config: dict,
        context: str = None,
        n_results=5,
        lang: str = "NotImplemented",
        silent_mode=False,
    ):
        """Find and parse the video"""

        json = cls._find(
            keywords=keywords,
            config=config,
            context=context,
            lang=lang,
            silent_mode=silent_mode,
        )
        videos = cls._parse(
            json,
            n_results=n_results,
            silent_mode=silent_mode,
        )

        return videos



"""
IO module
"""

import os

from awdible.logger import logger


class Io:
    """Class for input/output operations"""

    @classmethod
    def build_folders(
        cls,
        folders: list,
        create=True,
        silent_mode=True,
    ) -> list:
        """Check / Build the folders"""

        folders = [folders] if isinstance(folders, str) else folders

        # check if the folders are available
        for folder in folders:
            if not os.path.isdir(folder):
                logger.error(f"File not found : {folder}")
                if not silent_mode:
                    raise FileNotFoundError(f"File not found : {folder}")

                if create:
                    os.makedirs(folder)
                    logger.warning(f"Folder created : {folder}")

    @classmethod
    def build_files(
        cls,
        files: list,
        create=False,
        silent_mode=True,
    ) -> list:
        """Check / Build the files"""

        files = [files] if isinstance(files, str) else files

        # check if the files are available
        for file in files:
            if not os.path.isfile(file):
                logger.error(f"File not found : {file}")
                if not silent_mode:
                    raise FileNotFoundError(f"File not found : {file}")

                if create:
                    with open(file, "w") as f:
                        f.write("")
                    logger.warning(f"File created : {file}")

    @classmethod
    def clean_list_videos(
        cls,
        fn,
        rewrite=True,
        create=False,
        sort=True,
        silent_mode=False,
    ):
        """Clean the list of videos"""

        # TODO : this function is so broken!
        # TODO : an issue is opened for this
        # TODO : this issue is a bug and should be fixed asap

        # build the file if needed
        cls.build_files(
            fn,
            create=create,  # sould be false
            silent_mode=silent_mode,  # sould be true
        )

        # load the file
        with open(fn, "r") as f:
            lines = f.readlines()

        # clean the file
        lines = [line for line in lines if line]
        lines = [line.strip() for line in lines if line]
        lines = list(set(lines))

        # sort the file if needed
        lines = sorted(lines) if sort else lines

        # sep of hastag and not hastag
        not_hastag = [line for line in lines if not line.startswith("#")]
        hastag = [line for line in lines if not line.startswith("#")]

        # final
        final = not_hastag + hastag
        # final = [i for i in final if i.strip()]
        # final = list(set(final))
        final = [f"{line}\n" for line in final]

        # rewrite the file
        if rewrite:
            with open(fn, "w") as f:
                f.writelines(final)

        return hastag



"""
Convert module 
"""

import os
import subprocess

from awdible.logger import logger


class Convert:
    """Convert the video in various formats"""

    @classmethod
    def to_mp3(
        cls,
        src: str,
        overwrite: bool = False,
        remove_src: bool = True,
        silent_mode: False = False,
    ) -> str:
        """Convert the video to mp3"""

        if not os.path.exists(src):
            logger.error(f"File not found: {src}")
            if not silent_mode:
                raise FileNotFoundError(f"File not found: {src}")

        # get the destination
        dest = src.split(".")[:-1]
        dest = ".".join(dest) + ".mp3"

        logger.info(f"Converting  src : {src} to mp3 with dest : {dest}")

        # creating cmd
        cmd = f"ffmpeg -i {src} -q:a 0 -loglevel error -map a {dest}"
        if overwrite:
            cmd += " -y"

        # run the command
        logger.info(f"Running the command : cmd : {cmd}")
        prc = subprocess.run(cmd.split(" "), capture_output=True)

        # manage the output
        if prc.returncode:
            logger.warning(
                f"Subprocess  : code : {prc.returncode}, stdout : {prc.stdout} = > stderr : {prc.stderr}"
            )
            # raise an error if not silent mode
            if not silent_mode:
                raise ValueError(f"Error in the conversion : {prc.stderr}")

        logger.info(f"Conversion successful : {prc.stdout}")

        if remove_src:
            logger.info(f"Removing the source file : {src}")
            os.remove(src)

        return dest



"""
Pre-requirements module
"""

import subprocess

import requests

from awdible.logger import logger


class Prerequires:
    """Check the prerequires such as connection and ffmpeg availability"""

    @classmethod
    def has_connection(cls, silent_mode=False):
        """Check if the connection is available"""

        try:
            resp = requests.get("https://www.google.com")
            resp.raise_for_status()
            logger.info("Connection available")
            return True

        except Exception as e:
            logger.error(f"Error: {e}")
            logger.warning("Connection not available")
            if not silent_mode:
                raise e
            return False

    @classmethod
    def has_ffmpeg(cls, silent_mode=False):
        """Check if ffmpeg is installed"""

        try:
            cmd = "ffmpeg -version"
            prc = subprocess.run(cmd.split(" "), capture_output=True)

        except Exception as e:
            logger.error(f"Error: {e}")
            if not silent_mode:
                raise e
            return False

        if prc.returncode:
            logger.error(f"Error: {prc.stderr}")
            logger.error(
                f"Error: returncode : {prc.returncode}, stdout : {prc.stdout}, stderr : {prc.stderr}"
            )
            return False

        return True



"""
Module for validators and decorators 
"""

import os

from awdible.logger import logger
from awdible.validators import Validator

# class Number(Validator):
#     """Validator for numbers"""

#     def validate(self, value):
#         if not isinstance(value, (int, float)):
#             raise TypeError(
#                 f"Expected {self.private_name[1:]} to be an int or float : received {type(value)} for {value}"
#             )


# class Crop(Validator):
#     """Validator for numbers"""

#     MIN = 0
#     MAX = 3_600 * 10

#     def validate(self, value):
#         if not isinstance(value, (int, float, str)):
#             raise TypeError(
#                 f"Expected {self.private_name[1:]} to be an int or float  or str: received {type(value)} for {value}"
#             )

#         try:
#             value = int(value)
#         except Exception as e:
#             raise ValueError(
#                 f"Expected {self.private_name[1:]} to be an int : received {type(value)} for {value} : {e}"
#             )

#         if value < self.MIN or value > self.MAX:
#             raise ValueError(
#                 f"Expected {self.private_name[1:]} to be between {self.MIN} and {self.MAX} : received {value}"
#             )


# class Bool(Validator):
#     """Validator for bools"""

#     def validate(self, value):
#         if not isinstance(value, (bool, int)):
#             raise TypeError(
#                 f"Expected {self.private_name[1:]} to be a bool : received {type(value)} for {value}"
#             )
#         # try:
#         #     bool(value)
#         # except Exception as e:
#         #     raise ValueError(
#         #         f"Impossible to cast {self.private_name[1:]} to bool received {type(value)} for {value} : {e}"
#         #     )


# class File(Validator):
#     """Validator for file paths"""

#     def validate(self, value):
#         value = value if value else ""
#         if not isinstance(value, str):
#             raise TypeError(
#                 f"Expected {self.private_name[1:]} to be a str : received {type(value)} for {value}"
#             )

#         # if value and (not os.path.isfile(value)):
#         #     raise ValueError(
#         #         f"Expected {self.private_name[1:]} to be a valid file path : received {value}"
#         #     )


# class Dir(Validator):
#     """Validator for directory paths"""

#     def validate(self, value):
#         if not isinstance(value, str):
#             raise TypeError(
#                 f"Expected {self.private_name[1:]} to be a str : received {type(value)} for {value}"
#             )

#         # if not os.path.isdir(value):
#         #     raise ValueError(
#         #         f"Expected {self.private_name[1:]} to be a valid directory path : received {value}"
#         #     )


# class Output(Validator):
#     """Validator for output file format"""

#     def validate(self, value):
#         if not isinstance(value, str):
#             raise TypeError(
#                 f"Expected {self.private_name[1:]} to be a str : received {type(value)} for {value}"
#             )

#         # if value not in ["mp3", "mp4"]:
#         #     raise ValueError(
#         #         f"Expected {self.private_name[1:]} to be a valid file format : received {value}"
#         #     )


# class Context(Validator):
#     """Validator for context"""

#     def validate(self, value):
#         if not isinstance(value, str):
#             raise TypeError(
#                 f"Expected {self.private_name[1:]} to be a str : received {type(value)} for {value}"
#             )

#         value = value.lower()

#         # if value not in ["fr", "en"]:
#         #     raise ValueError(
#         #         f"Expected {self.private_name[1:]} to be a valid context : received {value}"
#         #     )


# class Config(Validator):
#     """Validator for config"""

#     DEFAULT_CONFIG = DEFAULT_CONFIG

#     def validate(self, value):
#         if not isinstance(value, dict):
#             raise TypeError(
#                 f"Expected {self.private_name[1:]} to be a dict : received {type(value)} for {value}"
#             )

#         for k, v in self.DEFAULT_CONFIG.items():
#             if k not in value.keys():
#                 logger.warning(f"Missing {k} in config. Using default value {v}")
#                 raise ValueError(
#                     f"Expected {self.private_name[1:]} to be have key {k} : {v}"
#                 )



"""
Awdible module
"""

import time


from ..config import config
from ..logger import logger

from .convert import Convert
from .crop import Crop
from .io import Io
from .prerequires import Prerequires
from .search import Search
from .video import Video
from .defaults import (
    DEFAULT_ASYNCHRONOUS,
    DEFAULT_CONFIG,
    DEFAULT_CONTEXT,
    DEFAULT_CROP_LIMIT,
    DEFAULT_DEST,
    DEFAULT_FILE,
    DEFAULT_FORCE,
    DEFAULT_LOG,
    DEFAULT_OUTPUT,
    DEFAULT_PORT,
    DEFAULT_PREFIX,
    DEFAULT_SEARCH,
    DEFAULT_SEARCH_NUMBER,
    DEFAULT_SLEEPER,
    DEFAULT_STREAMLIT,
    DEFAULT_TEST_MODE,
    DEFAULT_TMP,
    DEFAULT_VIDEO_ID,
    DEFAULT_VIDEO_URL,
    MAX_SEARCH_NUMBER,
    VIDEO_PREFIX,
)


class Awdible:
    """
    Awidble is a package that allows you to download the audio from a youtube video.
    It can be used to download the audio from a single video or from a list of videos.
    It can also be used to download the audio from a video that is longer than a certain duration.
    It can also be used to download the audio from a video that is the result of a search query.

    ----

    The Awdible class is the core of the awdible package.

    Agrs :

        video : str | list | None :
            The video url / list of videos to download from youtube. If you want to input a search query, use the --search option
            This attribute can be null only if the search option is set to True.

        dest : str :
            The destination directory of the file to download

        file : str | None :
            The input filewWith list of urls/videos. Default is None.
            If set, the video argument will be ignored.
            If not set and video not set, an error sound be raised.

        output : str :
            The output format of the file to download. Default is mp3.
            No other format is supported for now.

        crop_limit : int :
            The limit of duration of the video to download. If the video is longer than the limit, it will be cropped into smaller parts.
            This feature is not yet implemented.

        force : bool :
            If set, the download will be forced even if the ffmpeg is not installed.
            If not set and ffmpeg is not installed, an error will be raised.

        context: str :
            The language context of the video to download. Default is "en".
            This feature is not yet implemented.
            DEPRECIATION WARNING : This feature will be changed in the future.
            Context will be "song" / "video" "audiobook" "podcast" "live" and "en" "fr" "es" or None

        lang: str :
            This feature is not yet implemented.
            DEPRECIATION WARNING : This feature will be overlap the context feature in the future.

        search : bool :
            If set, the video argument will be treated as a search query and the first result will be downloaded.

        search_number : int :
            The number of results to download from the search query, if the search option is set to True.
            ie : search_number=5 will download the first 5 results from the search query.
            default is 1.
            MAX_SEARCH_NUMBER = 5

        prefix : bool :
            If set, can ommit the url prefix 'https://www.youtube.com/watch?v=' when passing the video url.
            DEPRECIATION WARNING : This feature will be removed in the future.

        asynchronous : bool :
            If set, the download will be asynchronous.
            This feature is not yet implemented.

        streamlit : bool :
            If set, a streamlit session should be launched.

        port : int :
            The port of the streamlit web app.

        test_mode : bool :
            If set, the test mode will be activated.

        config : dict :
            The config of the awdible package.
            Most important is the RAPID_API_KEY and RAPID_API_HOST for the youtube search allowed with the --search option.

        sleeper : int :
            The time to sleep between two downloads.
    """

    VIDEO_PREFIX = VIDEO_PREFIX

    DEFAULT_VIDEO_URL = DEFAULT_VIDEO_URL
    DEFAULT_VIDEO_ID = DEFAULT_VIDEO_ID

    DEFAULT_DEST = DEFAULT_DEST
    DEFAULT_FILE = DEFAULT_FILE
    DEFAULT_OUTPUT = DEFAULT_OUTPUT
    DEFAULT_CROP_LIMIT = DEFAULT_CROP_LIMIT
    DEFAULT_CONTEXT = DEFAULT_CONTEXT

    DEFAULT_SEARCH = DEFAULT_SEARCH
    DEFAULT_SEARCH_NUMBER = DEFAULT_SEARCH_NUMBER
    MAX_SEARCH_NUMBER = MAX_SEARCH_NUMBER

    DEFAULT_ASYNCHRONOUS = DEFAULT_ASYNCHRONOUS
    DEFAULT_PREFIX = DEFAULT_PREFIX

    DEFAULT_FORCE = DEFAULT_FORCE
    DEFAULT_TMP = DEFAULT_TMP
    DEFAULT_LOG = DEFAULT_LOG

    DEFAULT_STREAMLIT = DEFAULT_STREAMLIT
    DEFAULT_PORT = DEFAULT_PORT

    DEFAULT_TEST_MODE = DEFAULT_TEST_MODE

    DEFAULT_CONFIG = DEFAULT_CONFIG

    DEFAULT_SLEEPER = DEFAULT_SLEEPER

    def __init__(
        self,
        video: str | list = DEFAULT_VIDEO_URL,
        dest: str = DEFAULT_DEST,
        file: str = DEFAULT_FILE,
        output: str = DEFAULT_OUTPUT,
        force: bool = DEFAULT_FORCE,
        crop_limit: int = DEFAULT_CROP_LIMIT,
        context: str = DEFAULT_CONTEXT,
        lang: str = "NotImplemented",
        search: bool = DEFAULT_SEARCH,
        search_number: int = DEFAULT_SEARCH_NUMBER,
        prefix: bool = DEFAULT_PREFIX,
        asynchronous: bool = DEFAULT_ASYNCHRONOUS,
        streamlit: bool = DEFAULT_STREAMLIT,
        port: int = DEFAULT_PORT,
        test_mode: bool = DEFAULT_TEST_MODE,
        config: dict = config,
        sleeper=DEFAULT_SLEEPER,
    ):
        """Init the Awdible class"""

        self._awdible = None

        self.video = video
        _video_list = video if isinstance(video, list) else [video]
        self.video_list = list(set(_video_list))

        self.ok_video_list = []
        self.ko_video_list = []

        self.dest = dest
        self.file = file
        self.output = output
        self.crop_limit = crop_limit
        self.force = force
        # TODO : Change context to "song" / "video" "audiobook" "podcast" "live"
        # TODO : enabla lang isntead of context en "fr" "es" or None
        # TODO add auto context and lang detection
        self.context = context
        self.lang = lang

        self.search = search
        self.search_number = search_number

        # TODO ADD A VALIDATOR
        if self.search_number > self.MAX_SEARCH_NUMBER:
            self.search_number = self.MAX_SEARCH_NUMBER
        self.prefix = prefix
        self.asynchronous = asynchronous

        self.streamlit = streamlit
        self.port = port

        self.test_mode = test_mode

        self.config = config

        self.log = self.DEFAULT_LOG
        self.tmp = self.DEFAULT_TMP

        self.sleeper = sleeper

        self.ffmpeg_installed = Prerequires.has_ffmpeg()
        self.connection = Prerequires.has_connection()

        logger.warning(f"\n\n\n---NEW INIT---\n\n\n")

    def _pre_run(self):
        """Perform the pre run checks"""

        # manage if not ffmpeg installed
        if (not self.ffmpeg_installed) and self.force:
            logger.warning("ffmpeg is not installed BUT force option activated")

        if (not self.ffmpeg_installed) and (not self.force):
            logger.error(f"ffmpeg is not installed and --force option not activated")
            logger.error(f"consider installing ffmpeg or using --force option")
            raise AttributeError(
                "ffmpeg is not installed => Install ffmpeg or use --force option"
            )

        # manage if not connection
        if not self.connection:
            raise ConnectionError("No internet connection found")

        # manage if not not folders
        Io.build_folders([self.dest, self.tmp, self.log], create=True)

        # manage if file
        if self.file:
            self.video_list = Io.clean_list_videos(
                self.file,
                sort=True,
                rewrite=True,
            )

    def run(self):
        """Run the awdible session"""

        # pre run checks
        self._pre_run()

        # manage synch vs asynch
        if self.asynchronous:
            outs = self.run_asynch()
        else:
            outs = self.run_synch()

        logger.info(f"Outs : {outs}")

    def run_synch(self):
        """Run the awdible session synchronously"""

        outs = []

        # loop over the video list
        for video in self.video_list:
            # run one
            out = self.run_one_synch(video)
            outs.append(out)

            # add to relevant list
            if out:
                self.ok_video_list.append([video, out])
            else:
                self.ko_video_list.append([video, out])

            # sleep
            time.sleep(self.sleeper)

        outs = self._outro(outs)
        return outs

    def _outro(self, outs=None):
        """Outro of the awdible session"""

        logger.info(f"Video list: {self.video_list}")
        logger.warning(f"OK video list: {self.ok_video_list}")
        logger.warning(f"KO video list: {self.ko_video_list}")

        # TODO : add an io method to clean dest from .mp4 ONLY if 2 different files type
        # if only mp4, keep the files => maybe ffmpeg is not installed and --force option activated

        return outs

    async def run_asynch(self):
        """Run the awdible session asynchronously"""

        raise NotImplementedError("Sorry Bro! ")

    def run_one_synch(self, video: str):
        """Run the awdible session for one video"""

        # good case
        if video.startswith(self.VIDEO_PREFIX):
            try:
                fn = self._get_stream_save_convert(video)
                return fn
            except Exception as e:
                logger.error(f"Error: {e}")
                # self._outro()
                return None

        # if not valid and search is True:
        elif (not video.startswith(self.VIDEO_PREFIX)) and self.search:
            # get urls from search
            urls = self._find_parse(video, n_results=5)

            # manage if no urls
            logger.info(f"Urls found : {urls}")
            if not urls:
                return None

            # TODO to really enable this possibility of dowloading the first n results
            # TODO of the search we need, at least, to have a return [fn]
            # TODO and to manage the run_one out management of our attributes

            for url in urls[: self.search_number]:
                logger.info(f"Trying to download from url : {url}")
                try:
                    fn = self._get_stream_save_convert(url)
                    return fn
                except Exception as e:
                    logger.error(f"Error: {e}")
                    time.sleep(3)

            return None

        # if not valid and search is True:
        elif (not video.startswith(self.VIDEO_PREFIX)) and self.prefix:
            VIDEO_PREFIX = (
                self.VIDEO_PREFIX
                if self.VIDEO_PREFIX.endswith("/")
                else self.VIDEO_PREFIX + "/"
            )
            video = VIDEO_PREFIX + video
            try:
                fn = self._get_stream_save_convert(video)
                return fn
            except Exception as e:
                logger.error(f"Error: {e}")
                # self._outro()
                return None

        else:  # and self.default_prefix
            raise AttributeError("Sorry Bro! ")

    def _get_stream_save(self, url):
        """Get the audio from the video and save it"""

        fn = Video.get_stream_save(url, self.dest)

        return fn

    def _get_stream_save_convert(self, url):
        """Get the audio from the video and save it"""

        fn = Video.get_stream_save(url, self.dest)
        fn = self._convert(fn)

        return fn

    def _find_parse(
        self,
        keywords: str,
        n_results=5,
        retry=5,
    ) -> list[str]:
        """Find and parse the video"""

        urls = []

        logger.warning(f"Searching for video with keywords : {keywords}")

        for i in range(retry):
            try:
                urls = Search.find_parse(
                    keywords,
                    config=self.config,
                    context=self.context,
                    n_results=n_results,
                    lang=self.lang,
                    silent_mode=False,
                )

                logger.info(f"Urls found : {urls}")

                if urls:
                    return urls

            except Exception as e:
                logger.error(f"Error: {e}")

            time.sleep(3)
            logger.warning(
                f"Retry... {i+1} / {retry} for keywords : {keywords}, context : {self.context}, lang : {self.lang}"
            )

        return []

    def _convert(
        self,
        src: str,
        overwrite=True,
        remove_src=True,
    ) -> str:
        """Convert the video to mp3"""

        # convert
        fn = Convert.to_mp3(
            src,
            overwrite=overwrite,
            remove_src=remove_src,
        )

        return fn
